@@ -1,11 +1,100 @@
 import os
 import xml.etree.ElementTree as et
 import logging
+import pickle
+
+class Rule():
+    """Generic tokenization rule for ad hoc model creation."""
+
+    def __init__(self, action, value, key):
+        self.action = action
+        self.value = value
+        self.key = key
+
+    def decode(self):
+        return '%s\t%s\t%s\n' % (self.action, self.key, self.value)
+
+class SplitToken(Rule):
+    """Instruction to split token."""
+
+    def __init__(self, token, where):
+        assert where.replace('l', '').replace('m', '').replace('r', '') == '', '`where` parameter can only include characters `l`, `m`, or `r`'
+        super().__init__('s', token, where)
+
+class ReplaceToken(Rule):
+    """Instruction to replace token."""
+
+    def __init__(self, replace_from, replace_to):
+        super().__init__('r', replace_from, replace_to)
+
+class ReplaceCharacter(Rule):
+    """Instruction to replace character."""
+
+    def __init__(self, replace_from, replace_to):
+        super().__init__('c', replace_from, replace_to)
+
+class Model():
+    """This class represents set of tokenization rules for ad hoc model creation."""
+
+    def __init__(self):
+        self.sdata = ''
+        self.case_sensitive = False
+        self.bypass = False
+        self.rules = dict()
+        self._replacements = {'r': dict(), 'c': dict()}
+
+    def __repr__(self):
+        if self.bypass:
+            return 'set\tbypass\t1\n'
+        self.sdata = 'set\tcs\t%d\n' % (1 if self.case_sensitive else 0)
+        for action in self.rules:
+            for key in self.rules[action]:
+                for value in self.rules[action][key]:
+                    self.sdata += '%s\t%s\t%s\n' % (action, key, value)
+        return self.sdata
+
+    def add_rule(self, rule):
+        """This method adds tokenization rule to model.
+
+        Args:
+            *rule* is Rule instance
+        """
+        if rule.action in self._replacements:
+            if rule.value in self._replacements[rule.action]:
+                obsolete_rule = Rule(rule.action, rule.value, self._replacements[rule.action][rule.value])
+                self.remove_rule(obsolete_rule)
+        if rule.action in self.rules:
+            if rule.value in self.rules[rule.action]:
+                if rule.key in self.rules[rule.action][rule.value]:
+                    obsolete_rule = Rule(rule.action, rule.key, rule.value)
+                    self.remove_rule(obsolete_rule)
+        if rule.action not in self.rules:
+            self.rules[rule.action] = dict()
+        if rule.key not in self.rules[rule.action]:
+            self.rules[rule.action][rule.key] = set()
+        self.rules[rule.action][rule.key].add(rule.value)
+        if rule.action in self._replacements:
+            self._replacements[rule.action][rule.value] = rule.key
+
+    def remove_rule(self, rule):
+        """This method removes tokenization rule from model.
+
+        Args:
+            *rule* is Rule instance
+        """
+        if rule.action in self.rules:
+            if rule.key in self.rules[rule.action]:
+                if rule.value in self.rules[rule.action][rule.key]:
+                    self.rules[rule.action][rule.key].remove(rule.value)
+                    if len(self.rules[rule.action][rule.key]) == 0:
+                        del self.rules[rule.action][rule.key]
+                    if len(self.rules[rule.action]) == 0:
+                        del self.rules[rule.action]
 
 class Normalizer():
     """This class includes functions and methods for normalizing strings."""
 
-    def __init__(self, filename, tokenizer_name='', debug_mode=False, verbose_mode=False):
+    def __init__(self, tokenizer_name='', debug_mode=False, verbose_mode=False):
         logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s')
         self.debug = debug_mode
         self.verbose = verbose_mode
@@ -14,7 +103,6 @@ class Normalizer():
             logging.root.setLevel(logging.INFO)
         if self.debug:
             logging.root.setLevel(logging.DEBUG)
-        self.filename = filename
         self.tokenizer_name = tokenizer_name
         self.normalizer_result = {'original': '', 'normalized': '', 'map': [], 'r_map': []}
         self.content = dict()
@@ -39,6 +127,49 @@ class Normalizer():
     def data(self, obj):
         self.content = obj
 
+    def expand_instruction(self, g, node, visited):
+        """Helper function that traverses a path and returns terminal node.
+        For a directed graph *g*, it is assumed that each node has at most 1 descendant.
+
+        Args:
+            *g* is a dict(str, str) representing a graph
+            *node* is starting node
+            *visited* is a set to keep visited nodes for cycle detection
+        """
+        if node in visited:
+            raise RecursionError('Circular reference in replacement instruction regarding "%s"' % (node))
+        visited.add(node)
+        if node in g:
+            node = self.expand_instruction(g, g[node], visited)
+        return node
+
+    def merge_replacements(self, sdata):
+        """This function takes *sdata* config string, merges classes of "c" and "r" tokenization rules,
+        and returns corrected configuration string that accounts for transitive rules.
+
+        Args:
+            *sdata* is string with set of tokrules
+        """
+        ret = ''
+        replacements = dict()
+        for line in sdata.splitlines():
+            if not line.strip().startswith('#'):
+                [action, parameter, subject] = line.strip().split('\t')[0:3]
+                if action in ['c', 'r']:
+                    if action not in replacements:
+                        replacements[action] = dict()
+                    if subject not in replacements[action]:
+                        replacements[action][subject] = parameter
+                    elif replacements[action][subject] != parameter:
+                        raise ValueError('Conflicting instruction: (replace "%s" --> "%s") vs (replace "%s" --> "%s")' % (subject, replacements[action][subject], subject, parameter))
+                    continue
+            ret += '%s\n' % (line)
+        for action in replacements:
+            for node in replacements[action]:
+                replacements[action][node] = self.expand_instruction(replacements[action], node, set())
+                ret += '%s\t%s\t%s\n' % (action, replacements[action][node], node)
+        return ret
+
     def update_str_with_chmap(self, value, chmap):
         """This function zooms through a string *value*, replaces characters
         according to *chmap* dictionary object, and returns the updated string.
@@ -48,12 +179,13 @@ class Normalizer():
             updated += chmap[x] if x in chmap else x
         return updated
 
-    def make_tokenizer(self, sdata):
+    def make_tokenizer(self, sdata, update=False):
         """This function loads static set of tokienization rules stored in
         tab-delimited string into trie implemented as nested dictionary structure.
 
         Args:
             *sdata* is string with set of tokrules
+            *update*: when True, existing trie will be updated rather than new trie created
 
         Returns:
             The dictionary object to store the output
@@ -61,9 +193,9 @@ class Normalizer():
         Tokenization rules string requirements:
             1) each line represents single tokenization rule;
             2) each line has at least 3 columns:
-            `s`, `l`|`m`|`r` (or any combination of them), `word` == `split` `word` on the `left`|`middle`|`right`
+            `s`, `l`|`m`|`r` (or any combination of them), `word` == split `word` on the `left`|`middle`|`right`
             *or*
-            `r`, `another_word`, `word` == `replace` `word` with `another_word`
+            `r`, `another_word`, `word` == replace `word` with `another_word`
             *or*
             `c`, `another_char`, `char` == replace `char` with `another_char`
 
@@ -76,14 +208,18 @@ class Normalizer():
             '~r' => s r (split right)
         """
         # TODO: review for refacting, stay dry
+        sdata = self.merge_replacements(sdata)
         actions = {
             'r': {'': '~_'},
             's': {'l': '~l', 'm': '~m', 'r': '~r'},
             'c': {'': '~='}
         }
-        trie = dict()
-        trie['_settings'] = dict()
-        trie['_chmap'] = dict()
+        if update:
+            trie = self.content
+        else:
+            trie = dict()
+            trie['_settings'] = dict()
+            trie['_chmap'] = dict()
         for line in sdata.splitlines():
             if not line.strip().startswith('#'):
                 [action, parameter, subject] = line.strip().split('\t')[0:3]
@@ -142,6 +278,23 @@ class Normalizer():
             return 2
         return 0
 
+    def align_case(self, replacement, original, normalizer_option):
+        """This function aligns letter case in *replacement* with that in *original*
+        (only if *normalizer_option* assumes this is to be done)
+
+        Args:
+            *replacement* is string whose letter case is to be reviewed
+            *original* is string that is to serve as letter case template
+            *normalizer_option* is integer
+        """
+        if normalizer_option != 3:
+            return replacement
+        if original == original.upper():
+            return replacement.upper()
+        elif original[0] == original[0].upper() and original[1:] == original[1:].lower():
+            return replacement[0].upper() + replacement[1:].lower()
+        return replacement
+
     def reverse_map(self, m):
         """This function takes character location map in a form it is stored at self.normalizer_result['map']
         and returns list where item index is character index in original string, and item value is pair
@@ -172,7 +325,7 @@ class Normalizer():
         Args:
             *source_string* is input string to normalize
             *word_separator* is word separator to consider (must be single character)
-            *normalizer_option* is integer either 0 (normal, default), 1 (list), or 2 (set)
+            *normalizer_option* is integer either 0 (normal, default), 1 (list), 2 (set), 3 (join split tokens back)
         """
         assert len(word_separator) == 1, 'word_separator must be single character'
         # TODO: review for refactoring
@@ -180,6 +333,7 @@ class Normalizer():
         if source_string == '':
             return ''
         original_string = source_string
+        parsed_string = source_string
         subtrie = self.content
         if '_settings' in subtrie and 'bypass' in subtrie['_settings'] and subtrie['_settings']['bypass'] == '1':
             self.normalizer_result['normalized'] = original_string
@@ -187,7 +341,9 @@ class Normalizer():
             self.normalizer_result['r_map'] = [(i, i) for i in range(len(original_string))]
             return original_string
         if '_settings' not in subtrie or 'cs' not in subtrie['_settings'] or subtrie['_settings']['cs'] != '1':
-            original_string = original_string.lower()
+            parsed_string = parsed_string.lower()
+            if normalizer_option != 3:
+                original_string = parsed_string
         this_fragment = ''
         buffer = ''
         last_buffer = ''
@@ -196,9 +352,9 @@ class Normalizer():
         b_map = []
         l_map = []
         t_map = []
-        this_group = last_group = self.chargroup(original_string[0])
-        total_length = int(len(original_string))
-        character = original_string[0]
+        this_group = last_group = self.chargroup(parsed_string[0])
+        total_length = int(len(parsed_string))
+        character = parsed_string[0]
         last_character = ''
         current_index = 0
         temp_index = -1
@@ -207,10 +363,14 @@ class Normalizer():
         on_the_left = True
         on_the_right = False
         added_separator = False
+        separators = [0, 0]
+        last_separators = [0, 0]
+        separator_index = set()
         while current_index < total_length:
-            character = original_string[current_index]
+            character, original_character = parsed_string[current_index], original_string[current_index]
             if character in self.content['_chmap']:
                 character = self.content['_chmap'][character]
+                original_character = character
             this_group = self.chargroup(character)
             on_the_right = False
             added_separator = False
@@ -218,6 +378,7 @@ class Normalizer():
                 if this_group == 0 or this_group != last_group:
                     if not buffer.endswith(word_separator):
                         buffer += word_separator
+                        separators[1] = 1
                         if len(b_map) == len(buffer):
                             b_map[-1] = current_index
                         else:
@@ -232,6 +393,7 @@ class Normalizer():
                 if not began_reading:
                     if on_the_left and this_fragment and this_fragment[-1:] != word_separator:
                         this_fragment += word_separator
+                        separator_index.add(len(this_fragment) - 1)
                         if len(f_map) == len(this_fragment):
                             f_map[-1] = current_index
                         else:
@@ -240,13 +402,18 @@ class Normalizer():
                         f_map.pop()
                         this_fragment = this_fragment[:-1]
                     f_map += b_map
+                    if separators[0]:
+                        separator_index.add(len(this_fragment))
                     this_fragment += buffer
+                    if separators[1]:
+                        separator_index.add(len(this_fragment) - 1)
                     buffer = ''
+                    separators[0], separators[1] = 0, 0
                     b_map = []
                 on_the_left = on_the_left or added_separator or last_character == word_separator
                 began_reading = True
                 subtrie = subtrie[character]
-                buffer += character
+                buffer += original_character
                 b_map += [current_index for x in character]
             else:
                 on_the_right = on_the_right or character == word_separator
@@ -256,19 +423,23 @@ class Normalizer():
                 if '~_' in subtrie:
                     # we may need to apply this replacement in future, so keep buffer value and subtrie['~_']
                     last_buffer = buffer
-                    last_replacement = subtrie['~_']
+                    last_separators[0], last_separators[1] = separators[0], separators[1]
+                    last_replacement = self.align_case(subtrie['~_'], last_buffer, normalizer_option)
                     l_map = [b_map[0] for i in range(len(last_replacement))]
-                if '~_' in subtrie and ((on_the_left or on_the_right) or '~m' in subtrie):
+                if '~_' in subtrie and ((on_the_left and on_the_right) or '~m' in subtrie or ('~l' in subtrie and on_the_left) or ('~r' in subtrie and on_the_right)):
                     # now buffer has token to be replaced
-                    buffer = subtrie['~_'] + word_separator #if not buffer.endswith(word_separator) else ''
+                    buffer = self.align_case(subtrie['~_'], buffer, normalizer_option) + word_separator #if not buffer.endswith(word_separator) else ''
+                    separators[1] = 1 if parsed_string[current_index] != word_separator else 0
                     b_map = [b_map[0] for i in range(len(buffer))]
                     last_buffer = ''
+                    last_separators[0], last_separators[1] = 0, 0
                     l_map = []
                     temp_index = -1
                     # now buffer has replaced token
                 if '~l' in subtrie and on_the_left:
                     if not buffer.endswith(word_separator):
                         buffer += word_separator
+                        separators[1] = 1
                         if len(b_map) == len(buffer):
                             b_map[-1] = current_index
                         else:
@@ -277,20 +448,27 @@ class Normalizer():
                 if '~m' in subtrie and not on_the_left and not on_the_right:
                     if not buffer.startswith(word_separator):
                         buffer = word_separator + buffer
+                        separators[0] = 1
                         b_map.insert(0, current_index)
                     if not buffer.endswith(word_separator):
                         buffer += word_separator
+                        separators[1] = 1
                         if len(b_map) == len(buffer):
                             b_map[-1] = current_index
                         else:
                             b_map.append(current_index)
                     if last_buffer:
                         f_map = f_map[:-len(last_buffer)] + l_map
+                        if last_separators[0]:
+                            separator_index.add(len(this_fragment))
                         this_fragment = this_fragment[:-len(last_buffer)] + last_replacement
+                        if last_separators[1]:
+                            separator_index.add(len(this_fragment) - 1)
                     temp_index = -1
                 if '~r' in subtrie and on_the_right:
                     if not buffer.startswith(word_separator):
                         buffer = word_separator + buffer
+                        separators[0] = 1
                         b_map.insert(0, current_index)
                     temp_index = -1
                 subtrie = self.content
@@ -300,6 +478,7 @@ class Normalizer():
                     continue
                 if on_the_left and this_fragment and this_fragment[-1:] != word_separator and character != word_separator and not added_separator:
                     this_fragment += word_separator
+                    separator_index.add(len(this_fragment))
                     if len(f_map) == len(this_fragment):
                         f_map[-1] = current_index
                     else:
@@ -308,8 +487,13 @@ class Normalizer():
                     f_map.pop()
                     this_fragment = this_fragment[:-1]
                 f_map += b_map
+                if separators[0]:
+                    separator_index.add(len(this_fragment))
                 this_fragment += buffer
-                buffer = character
+                if separators[1]:
+                    separator_index.add(len(this_fragment) - 1)
+                buffer = original_character
+                separators[0], separators[1] = 0, 0
                 b_map = [current_index for x in character]
                 on_the_left = False
                 if character in self.content:
@@ -323,28 +507,40 @@ class Normalizer():
         # DRY!
         on_the_right = True
         on_the_left = this_fragment == '' or this_fragment[-1:] == word_separator
-        if '~_' in subtrie and ((on_the_left or on_the_right) or '~m' in subtrie):
+        if '~_' in subtrie and ((on_the_left and on_the_right) or '~m' in subtrie or ('~l' in subtrie and on_the_left) or ('~r' in subtrie and on_the_right)):
             # now buffer has token to be replaced
-            buffer = subtrie['~_'] + word_separator #if not buffer.endswith(word_separator) else ''
+            buffer = self.align_case(subtrie['~_'], buffer, normalizer_option) + word_separator #if not buffer.endswith(word_separator) else ''
+            separators[1] = 1
             b_map = [b_map[0] for i in range(len(buffer))]
             last_buffer = ''
+            last_separators[0], last_separators[1] = 0, 0
             l_map = []
             # now buffer has replaced token
         if '~r' in subtrie and on_the_right:
             if not buffer.startswith(word_separator):
                 buffer = word_separator + buffer
+                separators[0] = 1
                 b_map.insert(0, total_length - 1)
             if last_buffer:
                 f_map += l_map
+                if last_separators[0]:
+                    separator_index.add(len(this_fragment))
                 this_fragment = this_fragment[:-len(last_buffer)] + last_replacement
+                if last_separators[1]:
+                    separator_index.add(len(this_fragment))
         if on_the_left and this_fragment[-1:] != word_separator:
             this_fragment += word_separator
+            separator_index.add(len(this_fragment) + 1)
             f_map.append(total_length - 1)
         if this_fragment.endswith(word_separator) and buffer.startswith(word_separator):
             f_map.pop()
             this_fragment = this_fragment[:-1]
         f_map += b_map
+        if separators[0]:
+            separator_index.add(len(this_fragment))
         this_fragment += buffer
+        if last_separators[1]:
+            separator_index.add(len(this_fragment) - 1)
         while this_fragment.startswith(word_separator):
             this_fragment = this_fragment[len(word_separator):]
             f_map = f_map[len(word_separator):]
@@ -353,9 +549,11 @@ class Normalizer():
             f_map = f_map[:-len(word_separator)]
         normalized = this_fragment
         if normalizer_option == 1:
-            normalized = word_separator.join(sorted(normalized.split(word_separator)))
+            normalized = word_separator.join(sorted(this_fragment.split(word_separator)))
         elif normalizer_option == 2:
-            normalized = word_separator.join(sorted(set(normalized.split(word_separator))))
+            normalized = word_separator.join(sorted(set(this_fragment.split(word_separator))))
+        elif normalizer_option == 3:
+            normalized = ''.join([this_fragment[i] for i in range(len(this_fragment)) if i not in separator_index])
         elif len(f_map) > 0 and normalizer_option == 0:
             f_map[-1] = total_length - 1
             self.normalizer_result['map'] = f_map
@@ -363,6 +561,24 @@ class Normalizer():
         self.normalizer_result['normalized'] = normalized
         return normalized
 
+    def save(self, filename):
+        """This method pickles normalizer to a file.
+
+        Args:
+            *filename* is path/filename to save Normalizer object to
+        """
+        with open(filename, mode='wb') as f:
+            pickle.dump(self.data, f)
+
+    def load(self, filename):
+        """This function unpickles normalizer from a file.
+
+        Args:
+            *filename* is path/filename to load Normalizer object from
+        """
+        with open(filename, mode='rb') as f:
+            self.data = pickle.load(f)
+
 class Builder():
     """This class is the builder for Normalizer."""
 
@@ -376,7 +592,7 @@ class Builder():
             logging.basicConfig(level=logging.DEBUG)
 
     def wrap_result(self, root, address, keyhole, key, parent, child):
-        """This function populates dict object *keyhole* with the data retrieved from XML elemebt *root* as follows:
+        """This function populates dict object *keyhole* with the data retrieved from XML element *root* as follows:
             keyhole[key][root/address.col2][root/address.col3]
 
         Args:
@@ -437,21 +653,25 @@ class Builder():
         for key in ['set', 's', 'r', 'c']:
             for prop in data[key]:
                 for value in data[key][prop]:
-                    ret += '{0}\t{1}\t{2}\n'.format(key, prop, value)
+                    ret += '%s\t%s\t%s\n' % (key, prop, value)
         return (data['name'], ret)
 
-    def build_normalizer(self, filename=None):
+    def build_normalizer(self, endpoint=None):
         """This function loads configuration, constructs Normalizer with this configuration,
-        and returns this Normalizer object.
+        and returns created Normalizer object.
 
         Args:
-            str *filename* is XML file defining the configuration of a tokenizer
+            *endpoint* is either Model instance, or path to XML file defining the configuration of a tokenizer
         """
-        if not filename:
-            filename = '%s/tokenizer.standard.xml' % (os.path.abspath(os.path.dirname(__file__)))
-        (batch_name, data) = self.expose_tokenizer(filename)
-        machine = Normalizer(filename, batch_name)
+        if not endpoint:
+            endpoint = '%s/tokenizer.standard.xml' % (os.path.abspath(os.path.dirname(__file__)))
+        if isinstance(endpoint, Model):
+            batch_name, data = None, str(endpoint)
+            machine = Normalizer(batch_name, self.debug, self.verbose)
+        else:
+            (batch_name, data) = self.expose_tokenizer(endpoint)
+            machine = Normalizer(batch_name, self.debug, self.verbose)
         built = machine.make_tokenizer(data)
         if not built:
-            logging.critical('Could not build normalizer using "{0}"!'.format(filename))
+            logging.critical('Endpoint %s: Could not build normalizer' % (endpoint))
         return machine
